#!/usr/bin/env python3

# Copyright 2021 The Meson development team

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Hack to make relative imports to mlog possible
from pathlib import Path
import sys
root = Path(__file__).absolute().parents[1]
sys.path.insert(0, str(root))

# Now run the actual code
from refman.main import main

if __name__ == '__main__':
    raise SystemExit(main())
